
;PRINT USR stack,"a"[,"b"],find;
;find array 'a' [& optionally rename as 'b']; position returned in MEMBOT
;PRINT USR stack,"a","b",fill;
;fill array 'a' elements with value 'b'
;PRINT USR stack,a,b,l[,"c"],copy;
;[optionally copy string 'c' to 'a' then] copy 'l' bytes from 'a' to 'b' (may overlap)

;array name form: "a()" or "a$()"
;'a','b','c' may be any appropriate numeric or string expression

;copy may be used to fill with a recurring pattern; eg.
;PRINT USR stack,50000,50004,996,"wxyz",copy;
;will fill 50000-50999 with "wxyz" recurring
;If no initialisation string is defined then copy will start from
;the top or bottom end as appropriate, depending on the relative
;positions of 'a' & 'b', in order to preserve the source data at the
;destination in case the two areas overlap; ie. it effectively does
;a shift up or down as appropriate.

SCANNG: equ  $24fb           ;subroutine SCANNING
LOOKVA: equ  $28b2           ;subroutine LOOK_VARS
STORES: equ  $2ab6           ;calculator stack: from AEDCB
FETCHS: equ  $2bf1           ;calculator stack:   to AEDCB
STAKA:  equ  $2d28           ;calculator stack: from A
STAKBC: equ  $2d2b           ;calculator stack: from BC
FPTOBC: equ  $2da2           ;calculator stack:   to BC
FPTOA:  equ  $2dd5           ;calculator stack:   to A

CH_ADD: equ  23645
MEMBOT: equ  23698

LOCATE: equ  65000           ;***
        org  LOCATE


;initialisation: RANDOMIZE USR initsr
INITSR: ld   h,b
        ld   l,c        ;HL=runtime   location
        ld   de,LOCATE  ;DE=assembled location
        and  a
        sbc  hl,de
        push hl         ;HL=runtime displacement from LOCATE
        pop  bc         ;BC=runtime displacement from LOCATE
        ld   hl,FINDAR  ;--->adjust FINDAR call addresses
        add  hl,bc
        ex   de,hl
        ld   hl,CALL1+1
        add  hl,bc
        ld   (hl),e
        inc  hl
        ld   (hl),d     ;<---
        ld   hl,CALL2+1
        add  hl,bc
        ld   (hl),e
        inc  hl
        ld   (hl),d     ;<---
        ret


;stack function parameters passed from BASIC by:
;PRINT USR stack[,arg1,arg2...,argn],address;
STACK:  ld   hl,14
        add  hl,sp
        ld   sp,hl      ;undo USR
        ld   ix,0
GETARG: rst  $20
        call SCANNG
        inc  ix
        rst  $18
        cp   ","
        jr   z,GETARG
        call FPTOBC     ;BC=address
        ld   h,b
        ld   l,c
        push ix
        pop  bc
        dec  bc
        ld   a,c        ;A =number of parameters
        jp   (hl)       ;JP address


FINDAR: call FETCHS     ;DE=address, BC=length (of array name)
        ld   bc,(CH_ADD)
        push bc
        ld   (CH_ADD),de
        call LOOKVA     ;if found: Carry reset, HL=address; otherwise: Carry set
        ld   a,c        ;A =discriminator byte; bits 6&5 indicate variable type
        pop  bc
        ld   (CH_ADD),bc
        ret  c          ;return if not found
        push hl         ;HL=address of array
        inc  hl
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=1+ length of dims+elements
        inc  hl
        ld   b,(hl)     ;B =number of dimensions
        inc  hl
        dec  de
ADDUPD: inc  hl
        dec  de
        inc  hl
        dec  de
        djnz ADDUPD
        ld   b,d
        ld   c,e        ;BC=total size of array elements
        ex   de,hl      ;DE=address of array's 1st element
        pop  hl         ;HL=address of array
        and  a          ;reset Carry
        ret


;PRINT USR stack,"a"[,"b"],find;
;find array 'a' [& optionally rename as 'b']
FIND:   cp   1
        jr   z,CALL1-1
        call FETCHS     ;DE=address of new name
        ld   a,(de)
        push af
CALL1:  call FINDAR     ;***
        pop  bc
        jp   c,$1C2E    ;"2 Variable not found"
        push HL         ;HL=address of array
        ld   a,b
        cp   1
        jr   z,NOCHNG   ;jump if no change
        pop  HL
        cp   $61        ;compare "a"
        jr   c,CHECKA
        sub  $20
CHECKA: cp   $41        ;compare "A"
        jr   c,CHECK0
        sub  $40
CHECK0: cp   $1B
        jp   nc,$34E7   ;"A Invalid argument"
        ld   b,a
        ld   a,$E0
        and  (hl)
        add  a,b
        ld   (hl),a
        push hl
NOCHNG: pop  BC         ;BC=address of array
        ld   (MEMBOT),bc
        ret


;PRINT USR stack,"a","b",fill;
;fill array 'a' elements with value 'b'
FILL:   rst  $28
        defb $01,$38    ;exchange top two calculator stack items
CALL2:  call FINDAR     ;***
        jp   c,$1C2E    ;"2 Variable not found"
        push bc      ;1>;BC=total size of array elements
        push de      ;2>;DE=address of array's 1st element
        push af      ;3>;A =discriminator byte; bits 6&5 indicate variable type
        call FETCHS     ;get fill value parameters
        pop  hl      ;3<
        bit  6,h        ;H =discriminator byte
        jr   nz,FILL2   ;jump if character array
        ld   hl,MEMBOT+4
        ld   (hl),b
        dec  hl
        ld   (hl),c
        dec  hl
        ld   (hl),d
        dec  hl
        ld   (hl),e
        dec  hl
        ld   (hl),a
        ld   bc,5
        ex   de,hl
FILL2:  ex   de,hl      ;HL=fill value address
        pop  de      ;2<;DE=address of array's 1st element
        push de      ;2>
        push bc      ;3>
        ldir            ;copy fill value to 1st element
        pop  bc      ;3<;BC=size of fill value
        pop  af      ;2<;AF=address of array's 1st element
        pop  hl      ;1<;HL=total size of array elements
        push af      ;1>
        and  a
        sbc  hl,bc
        ld   b,h
        ld   c,l        ;BC=size of unfilled segment
        pop  hl      ;1<;HL=address of array's 1st element
        ldir
        ret


;PRINT USR stack,a,b,l[,"c"],copy;
;[optionally copy string 'c' to 'a' then] copy 'l' bytes from 'a' to 'b' (may overlap)
COPY:   cp   4
        jr   z,COPY2
        xor  a
        call STAKA      ;stack zero for an empty string
COPY2:  call FETCHS     ;get initialisation string
        push bc      ;1>;BC=string length
        push de      ;2>;DE=string address
        call FPTOBC
        pop  de      ;2<
        pop  hl      ;1<
        push bc      ;1>;BC=copy length
        push hl      ;2>
        push de      ;3>
        call FPTOBC
        pop  de      ;3<
        pop  hl      ;2<
        push bc      ;2>;BC=copy to address
        push hl      ;3>
        push de      ;4>
        call FPTOBC
        ld   d,b
        ld   e,c        ;DE=copy from address
        pop  hl      ;4<;HL=initialisation string address
        pop  bc      ;3<;BC=initialisation string length
        push de      ;3>
        ld   a,b
        or   c
        jr   z,COPY3
        ldir            ;copy initialisation string
        scf
COPY3:  pop  hl      ;3<
        pop  de      ;2<
        pop  bc      ;1<
        jr   c,LOWER    ;jump if init string set
CHUNK2: push hl
        sbc  hl,de      ;from - to
        pop  hl
        jr   c,HIGHER
LOWER:  ldir            ;from > to -or- initialisation string specified
        ret
HIGHER: add  hl,bc      ;from < to
        dec  hl         ;HL=end of from chunk
        push hl
        ex   de,hl
        add  hl,bc
        dec  hl
        ex   de,hl      ;DE=end of to chunk
        pop  hl
        lddr
        ret

LABELS0:defw INITSR
        defw STACK
        defw FIND
        defw FILL
        defw COPY